/*
############################################################################
# Copyright (c) 2012 Dave Hein
# MIT Licensed
############################################################################
*/
/*
  This object implements an EEPROM file system.  A sector size of 64
  is used, and the EEPROM must have a page size of 64 bytes or greater.
  Files consist of multiple sectors that a daisy-chained together.
  The first long of each sector points to the next sector in the
  chain.
 
  When the file system is first created the least significant
  bit of first long of each sector is set to one to indicate that
  the sector is unused.  When a sector is allocated for a file the
  last six bits in the first long are cleared to indicate that it
  is in use.
 
  The first long in the first sector contains the ASCII characters
  "EEFS" to indicate that this is a valid EEPROM file system.  The
  next long contains the address of the root directory, which is
  normally the second sector in the file system.  The third long
  contains the size of the root directory in bytes.
 
  In general, directory entries are 30 bytes in size.  They consist
  of 4 bytes for the starting address and attribute bits, 4 bytes
  for the file size and 22 bytes for the file name.  There are two
  30-byte directory entries per sector.  Since the file starting
  address is a multiple of 64, the 6 least significant bit are
  zero.  These bits are used to store the attibutes for the file.
  The attributes bits are directory, read, write and execute, or
  "drwx".
 
  A fifth bit is set to indicate that the directory entry is active
  and does not represent a deleted file.  If all 6 of the least
  significant bits are zero, the directory entry is inactive, and
  it is available to be used for a new file.
 
  A 64-byte cache is used to hold the contents of the current sector
  that is being read or written.  Two flags are used to indicate
  the range of bytes that have changed in the sector.  These are
  called dirtylow and dirtyhigh.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "eefs.h"
#include "pasm_i2c_driver.h"

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

// CONSTANTS
#define magic_word   0x45454653          // First word of a valid file system
#define sector_size  64                  // Number of bytes in a sector
#define device_sel   0xa0                // EEPROM I2C Device Select
#define attrib_mask  0x3f                // Mask for the file attribute bits
#define sector_mask  0xffffffc0          // Mask for the sector starting address
#define dirent_size  30                  // Number of bytes in a directory entry
#define fname_size   22                  // Maximum chars in a file name

// Local functions
static long UpdateFilesize(FileHandleT *handle, long position, long filesize);
static long WriteDirtyBlock(long sectoraddr, long *buffer, long *pdirtylow, long *pdirtyhigh);
static long ClearCached(FileHandleT *handle);
static void InitFileHandle(FileHandleT *handle, long directaddr, long firstaddr, long filesize, long *cacheptr);
static void InitRootHandle(FileHandleT *handle);
static long WriteEEPromLong(long addr, long value);
static long FindFreeSector(void);
static long FreeSectorChain(long sectoraddr);
static long FindFile(char *fname, FileHandleT *dirhand, long *pfilesize);
static long Remove(char *fname, int files_only);
static long CreateFile(char *fname, long modebits);
static long OpenRead(char *fname, FileHandleT *handle);
static long OpenAppend(char *fname, FileHandleT *handle);
static long OpenWrite(char *fname, FileHandleT *handle);
static long UpdateDirectoryEntry(FileHandleT *dirhand, FileHandleT *handle, long firstaddr, long filesize, char *fname);

// Local Variables
static long disk_start = 0x8000;
static long disk_end = 0;
static long eeprom_pin = 0;  // First pin of the EEPROM
static long first_free = 0x8000;
static FileHandleT local_handle;
static long readdir_block[8];  //attrib, size, fname

/*
  This routine initializes the disk for file access.  It returns
  a zero if the disk is ready, or a negative error code if it is not.
*/
long eefs_mount_explicit( long clkpin,  long datpin,  long startaddr,  long size)
{
  eeprom_pin = clkpin;
  disk_start = startaddr;
  first_free = disk_start + sector_size*2;
  pasm_i2c_driver_Initialize(eeprom_pin);
  if ((pasm_i2c_driver_ReadLong(eeprom_pin, device_sel, disk_start) != magic_word))
  {
    printf("EEPROM is not formatted.  Formating...\n");
    eefs_format(size);
  }
    //return -20 ' Signal mount failure
  disk_end = pasm_i2c_driver_ReadLong(eeprom_pin, device_sel, disk_start + 12);
  InitRootHandle(&local_handle);
  return 0;
}

/*
  This routine opens a file for access.  The access mode
  is determined by the first character of the string "mode".
  The allowable values for the first character are "r", "w"
  and "a" for read, write and append respectively.  The
  pass value of the handle pointer is return if successful.
*/
long eefs_hopen(FileHandleT *handle,  char *fname,  long mode)
{
  long result = 0;
  if (mode == 'd')
    result = Remove(fname, 1);
  else if ((mode == 'D'))
    result = Remove(fname, 0);
  else if ((handle == 0))
    return -EIO;
  else if ((mode == 'r'))
    result = OpenRead(fname, handle);
  else if ((mode == 'w'))
    result = OpenWrite(fname, handle);
  else if ((mode == 'a'))
    result = OpenAppend(fname, handle);
  return result;
}

long eefs_hclose(FileHandleT *handle)
{
  long position, filesize;
  position = handle->position;
  filesize = handle->filesize;
  eefs_hflush(handle);
  handle->cachedaddr = -1;
  UpdateFilesize(handle, position, filesize);
  return 0;
}

/*
  This routine reads "num" bytes from the file defined by "handle".
  The bytes are copied into the memory pointed to by "buffer".
  The return value consist of the number of bytes that were actually
  read.  A zero return value indicates an end of file or an error
  condition.
*/
long eefs_hread(FileHandleT *handle,  char *buffer,  long num)
{
  long num1;
  long sectoraddr = handle->sectoraddr;
  long offset = handle->offset;
  long position = handle->position;
  long filesize = handle->filesize;
  long cachedaddr = handle->cachedaddr;
  long *cacheptr = handle->cacheptr;
  long i = 0;
  
  while (i < num && position < filesize)
  {
    if ((offset >= sector_size))
    {
      if (cachedaddr != sectoraddr)
      {
        sectoraddr = pasm_i2c_driver_ReadLong(eeprom_pin, device_sel, sectoraddr);
      }
      else
      {
        sectoraddr = *cacheptr;
      }
      offset = 4;
    }
    if (cachedaddr != sectoraddr)
    {
      pasm_i2c_driver_ReadPage(eeprom_pin, device_sel, sectoraddr, (long)cacheptr, sector_size);
      cachedaddr = sectoraddr;
    }
    num1 = MIN((num - i), (filesize - position)); // Get smallest value of num - i,
    num1 = MIN(num1, (sector_size - offset)); // filesize - position, and sector_size - offset
    memcpy(buffer + i, (void *)cacheptr + offset, num1);
    position += num1;
    offset += num1;
    i += num1;
  }
  handle->sectoraddr = sectoraddr;
  handle->offset = offset;
  handle->position = position;
  handle->cachedaddr = cachedaddr;
  return i;
}

/*
  This routine writes "num" bytes to the file defined by "handle".
  The bytes are copied from the memory pointed to by "buffer".
  The return value consist of the number of bytes that were actually
  written.  A zero return value indicates an error occurred.
*/
long eefs_hwrite(FileHandleT *handle,  char *buffer,  long num)
{
  long result = 0;
  long *cacheptr;
  long sectoraddr, offset, position, filesize, cachedaddr, dirtylow, dirtyhigh, firstaddr, directaddr, num1, nextaddr;

  if (handle == 0) return -1;
  
  if (num < 1) return 0;

  result = num;
  sectoraddr = handle->sectoraddr;
  offset = handle->offset;
  position = handle->position;
  filesize = handle->filesize;
  cachedaddr = handle->cachedaddr;
  cacheptr = handle->cacheptr;
  dirtylow = handle->dirtylow;
  dirtyhigh = handle->dirtyhigh;
  firstaddr = handle->firstaddr;
  directaddr = handle->directaddr;

  if ((sectoraddr == 0))
  {
    sectoraddr = FindFreeSector();
    firstaddr |= sectoraddr;
    WriteEEPromLong(directaddr, firstaddr);
    cachedaddr = sectoraddr;
    *cacheptr = 0;
    dirtylow = 0;
    dirtyhigh = 4;
  }
  else if ( cachedaddr != sectoraddr)
  {
    WriteDirtyBlock(cachedaddr, cacheptr, &dirtylow, &dirtyhigh);
    pasm_i2c_driver_ReadPage(eeprom_pin, device_sel, sectoraddr, (long)cacheptr, sector_size);
    cachedaddr = sectoraddr;
  }
  while (1)
  {
    num1 = MIN(num, (sector_size - offset));
    memcpy((void *)cacheptr + offset, buffer, num1);
    dirtylow = MIN(dirtylow, offset); // Update dirtylow if offset is smaller
    num -= num1;
    position += num1;
    offset += num1;
    dirtyhigh = MAX(dirtyhigh, offset); // Update dirtyhigh if offset is larger
    if ((num <= 0))
    {
      break;
    }
    buffer += num1;
    nextaddr = *cacheptr;
    if (nextaddr)
    {
      WriteDirtyBlock(sectoraddr, cacheptr, &dirtylow, &dirtyhigh);
      pasm_i2c_driver_ReadPage(eeprom_pin, device_sel, nextaddr, (long)cacheptr, sector_size);
    }
    else
    {
      nextaddr = FindFreeSector();
      if (nextaddr == 0)
      {
        return 0;
      }
      *cacheptr = nextaddr;
      dirtylow = 0;
      dirtyhigh = MAX(dirtyhigh, 4); // Update dirtyhigh if less than 4
      WriteDirtyBlock(sectoraddr, cacheptr, &dirtylow, &dirtyhigh);
      *cacheptr = 0;
      dirtylow = 0;
      dirtyhigh = 4;
    }
    sectoraddr = nextaddr;
    cachedaddr = sectoraddr;
    offset = 4;
  }
  if (!dirtylow) // Write if sectoraddr has changed
  {
    WriteDirtyBlock(sectoraddr, cacheptr, &dirtylow, &dirtyhigh);
  }

  handle->sectoraddr = sectoraddr;
  handle->offset = offset;
  handle->position = position;
  handle->filesize = filesize;
  handle->cachedaddr = cachedaddr;
  handle->dirtylow = dirtylow;
  handle->dirtyhigh = dirtyhigh;
  handle->firstaddr = firstaddr;

  return result;
}

long eefs_hgetc(FileHandleT *handle)
{
  long val;
  long result;
  if ((result = eefs_hread(handle, (char *)&val, 1)) == 1)
  {
    result = val & 0xff;
  }
  return result;
}

long eefs_hputs(FileHandleT *handle,  char *buffer)
{
  return eefs_hwrite(handle, buffer, strlen(buffer));
}

long eefs_hputc(FileHandleT *handle,  long val)
{
  long result;
  if ((result = eefs_hwrite(handle, (char *)&val, 1)) == 1)
  {
    result = 0;
  }
  return result;
}

/*
  This routine opens the current directory for reading by readdir.
  A handle is returned, which is used by subsequent calls by readdir.
*/
long eefs_hopendir(FileHandleT *handle)
{
  long filesize = local_handle.filesize;
  long firstaddr = local_handle.firstaddr;
  long directaddr = local_handle.directaddr;
  InitFileHandle(handle, directaddr, firstaddr, filesize, handle->buffer);
  return 0;
}

/*
  This routine returns the size of the file in bytes.
*/
long eefs_hget_filesize(FileHandleT *handle)
{
  if (!handle) return 0;
  return handle->filesize;
}

/*
  This routine moves to the location specified by "newpos".
*/
long eefs_hseek(FileHandleT *handle,  long newpos)
{
  long sectoraddr, offset, position, filesize, num;
  if (!handle) return 0;
  sectoraddr = handle->sectoraddr;
  offset = handle->offset;
  position = handle->position;
  filesize = handle->filesize;
  if (newpos < 0 || newpos > filesize) return 0;
  // Check if the filesize needs to be updated before changing position
  UpdateFilesize(handle, position, filesize);
  // Backup to the beginning of the current block
  position -= offset - 4;
  // Backup to the beginning of the file if newpos < position
  if (newpos < position)
  {
    sectoraddr = handle->firstaddr & sector_mask;
    position = 0;
  }
  // Advance to the new position
  offset = newpos - position + 4;
  position = newpos;
  while (offset > sector_size)
  {
    sectoraddr = pasm_i2c_driver_ReadLong(eeprom_pin, device_sel, sectoraddr);
    offset -= sector_size - 4;
  }
  handle->sectoraddr = sectoraddr;
  handle->offset = offset;
  handle->position = position;
  return 1;
}

/*
  This routine returns the current position from the beginning of the file.
*/
long eefs_htell(FileHandleT *handle)
{
  if (!handle) return 0;
  return handle->position;
}

/*
  This routine writes out the data that is sitting in the write buffer.
*/
long eefs_hflush(FileHandleT *handle)
{
  long *cacheptr;
  long sectoraddr, cachedaddr, dirtylow, dirtyhigh;
  if (!handle) return 0;
  sectoraddr = handle->sectoraddr;
  cachedaddr = handle->cachedaddr;
  cacheptr = handle->cacheptr;
  dirtylow = handle->dirtylow;
  dirtyhigh = handle->dirtyhigh;
  WriteDirtyBlock(sectoraddr, cacheptr, &dirtylow, &dirtyhigh);
  handle->dirtylow = dirtylow;
  handle->dirtyhigh = dirtyhigh;
  return 0;
}

/*
  This routine returns the size of a file handle in bytes so that
  the calling routine can allocate a handle that is used by most
  of the file access routines.
*/
long eefs_handlesize(void)
{
  return sizeof(FileHandleT);
}

/*
  This routine formats the disk with an empty file system.
*/
long eefs_format( long size)
{
  long sectoraddr;
  disk_end = size;
  WriteEEPromLong(disk_start, magic_word);
  sectoraddr = disk_start + sector_size;
  WriteEEPromLong(disk_start + 4, sectoraddr | 1);
  WriteEEPromLong(disk_start + 8, 0);
  WriteEEPromLong(disk_start + 12, disk_end);
  WriteEEPromLong(sectoraddr, 0);
  sectoraddr += sector_size;
  first_free = sectoraddr;
  while ((sectoraddr < disk_end))
  {
    WriteEEPromLong(sectoraddr, 1);
    sectoraddr += sector_size;
  }
  InitRootHandle(&local_handle);
  return 0;
}

/*
  This routine renames a file in the files system.  It returns
  a value of zero if the file did not exist, or a one if it did.
*/
long eefs_rename(char *fname1,  char *fname2)
{
  long filesize;
  if (FindFile(fname2, &local_handle, &filesize)) return -1;
  if (!FindFile(fname1, &local_handle, &filesize)) return 0;
  local_handle.position -= fname_size;
  local_handle.offset -= fname_size;
  eefs_hwrite(&local_handle, fname2, fname_size);
  eefs_hflush(&local_handle);
  return 1;
}

/*
  This routine reads the information for the next file in the
  directory indicated by the handle "dirhand".  It returns a
  pointer to the file directory information block if the end of
  the directory has not been reached.  Otherwise it will return
  a zero.  The file directory information block contains 3 longs
  with the filesize, attributes, a physical size followed by the
  file name.
*/
long eefs_hreaddir(FileHandleT *dirhand)
{
  long sectoraddr, filesize, filename;
  if (dirhand == 0) return 0;
  while (1)
  {
    if (eefs_hread(dirhand, (char *)&sectoraddr, 4) != 4) return 0;
    if (eefs_hread(dirhand, (char *)&filesize, 4) != 4) return 0;
    if (eefs_hread(dirhand, (char *)&readdir_block[2], fname_size) != fname_size) return 0;
    //break; //Uncomment this line to display deleted files
    if (sectoraddr & attrib_mask) break;
  }
  if (!(sectoraddr & attrib_mask))
  {
    readdir_block[0] = 0;
    readdir_block[1] = 0;
  }
  else
  {
    readdir_block[0] = sectoraddr;
    readdir_block[1] = filesize;
  }
  return (long)&readdir_block;
}

/*
  This routines closes the directory that had previously been
  opened by opendir.
*/
long eefs_hclosedir(FileHandleT *dirhand)
{
  return 0;
}

/*
  This routine changes the current working directory to the path indicated by
  "path".  A value of zero is returned on success, and minus one is returned if
  the path does not exist.
*/
long eefs_hchdir(char *path)
{
  char *nextptr;
  long directaddr, sectoraddr, filesize;
  
  InitRootHandle(&local_handle); // Start from root directory
  if (*path == '/') path++;      // Skip initial '/' characer if present

  // Loop over sub-directory names in path
  while (*path)
  {
    nextptr = path;
    while (*nextptr)
    {
      if (*nextptr++ == '/')
      {
        nextptr[-1] = 0;
        break;
      }
    }
    sectoraddr = FindFile(path, &local_handle, &filesize);
    if (*nextptr) nextptr[-1] = '/'; // Restore '/'
    if (!sectoraddr) return -ENOENT; // Quit if subdirectory does not exist
    if (!(sectoraddr & ATTR_DIRECTORY)) return -ENOTDIR;
    directaddr = local_handle.sectoraddr + local_handle.offset - dirent_size;
    InitFileHandle(&local_handle, directaddr, sectoraddr, filesize, local_handle.buffer);
    path = nextptr;
  }
  return 0;
}

long eefs_hmkdir(char *fname,  long mode)
{
  return CreateFile(fname, 0xf) - 1;
}

long eefs_chmod(char *fname,  long modebits)
{
  long sectoraddr, filesize;
  if (!(sectoraddr = FindFile(fname, &local_handle, &filesize))) return 0;
  local_handle.position -= dirent_size;
  local_handle.offset -= dirent_size;
  sectoraddr = (sectoraddr &sector_mask) | 0x10 | (modebits &0xf);
  eefs_hwrite(&local_handle, (char *)&sectoraddr, 4);
  eefs_hflush(&local_handle);
  return 0;
}

// ****************************************************************
// Private Functions
// ****************************************************************
static long UpdateFilesize(FileHandleT *handle,  long position,  long filesize)
{
  long directaddr;
  if (position > filesize)
  {
    filesize = position;
    handle->filesize = filesize;
    directaddr = handle->directaddr;
    WriteEEPromLong(directaddr + 4, filesize);
  }
  return 0;
}

static long WriteDirtyBlock(long sectoraddr,  long *buffer,  long *pdirtylow,  long *pdirtyhigh)
{
  long dirtylow = *pdirtylow;
  long dirtyhigh = *pdirtyhigh;
  long num = dirtyhigh - dirtylow;
  long eepaddr = sectoraddr + dirtylow;
  long hubaddr = (long)buffer + dirtylow;
  if (dirtylow < 0 || dirtyhigh > sector_size) return 0;
  if (num <= 0) return 0;
  pasm_i2c_driver_WritePage(eeprom_pin, device_sel, eepaddr, hubaddr, num);
  while (pasm_i2c_driver_WriteWait(eeprom_pin, device_sel, eepaddr));
  *pdirtylow = sector_size;
  *pdirtyhigh = 0;
  return 0;
}

static long ClearCached(FileHandleT *handle)
{
  handle->cachedaddr = -1;
  handle->dirtylow = sector_size;
  handle->dirtyhigh = 0;
  return 0;
}

/*
  This routine intializes a file handle with the directory location
  and file information for a file.  The handle is initialized for
  the beginning of the file.
*/
static void InitFileHandle(FileHandleT *handle,  long directaddr,  long firstaddr,  long filesize,  long *cacheptr)
{
  handle->directaddr = directaddr;
  handle->firstaddr  = firstaddr;
  handle->sectoraddr = firstaddr & sector_mask;
  handle->offset     = 4;
  handle->filesize   = filesize;
  handle->position   = 0;
  handle->cachedaddr = -1;
  handle->cacheptr   = cacheptr;
  handle->dirtylow   = sector_size;
  handle->dirtyhigh  = 0;
}

/*
  This routine initializes a file handle for to point to the beginning
  of the root directory.
*/
static void InitRootHandle(FileHandleT *handle)
{
  long firstaddr, filesize;
  firstaddr = pasm_i2c_driver_ReadLong(eeprom_pin, device_sel, disk_start + 4); // Get root directory address
  filesize = pasm_i2c_driver_ReadLong(eeprom_pin, device_sel, disk_start + 8); // Get root directory size
  InitFileHandle(handle, disk_start + 4, firstaddr, filesize, handle->buffer);
}

// This routine writes one long to the EEPROM and waits till done
static long WriteEEPromLong( long addr,  long value)
{
  pasm_i2c_driver_WriteLong(eeprom_pin, device_sel, addr, value);
  while (pasm_i2c_driver_WriteWait(eeprom_pin, device_sel, addr));
  return 0;
}

// This routine locates a free sector
static long FindFreeSector(void)
{
  long sectoraddr;
  sectoraddr = first_free;
  while (sectoraddr < disk_end)
  {
    if ((pasm_i2c_driver_ReadByte(eeprom_pin, device_sel, sectoraddr) & 1))
    {
      first_free = sectoraddr + sector_size;
      return sectoraddr;
    }
    sectoraddr += sector_size;
  }
  return 0;
}

static long FreeSectorChain( long sectoraddr)
{
  long nextaddr;
  while (sectoraddr)
  {
    nextaddr = pasm_i2c_driver_ReadLong(eeprom_pin, device_sel, sectoraddr);
    WriteEEPromLong(sectoraddr, nextaddr | 1);
    first_free = MIN(first_free, sectoraddr);
    sectoraddr = nextaddr;
  }
  return 0;
}

// This routine searches for a file and returns its starting address
static long FindFile(char *fname, FileHandleT *dirhand,  long *pfilesize)
{
  char filename[24];
  long firstaddr, filesize;
  
  if (dirhand == 0) return 0;
  eefs_hseek(dirhand, 0);
  ClearCached(dirhand);
  while (1)
  {
    if (eefs_hread(dirhand, (char *)&firstaddr, 4) != 4) break;
    eefs_hread(dirhand, (char *)&filesize, 4);
    eefs_hread(dirhand, filename, fname_size);
    if (!strcmp(fname, filename) && (firstaddr & attrib_mask))
    {
      *pfilesize = filesize;
      return firstaddr;
    }
  }
  return 0;
}

/*
  This routine removes a file from the file system.  It returns
  a value of zero if the file did not exist, or a one if it did.
*/
static long Remove(char *fname, int files_only)
{
  long directaddr, sectoraddr, filesize;
  if (!(sectoraddr = FindFile(fname, &local_handle, &filesize)))
    return -ENOENT;
  if (files_only && (sectoraddr & ATTR_DIRECTORY))
    return -EISDIR;
  // Remove file from the directory by clearing it's attribute bits
  sectoraddr &= sector_mask;
  directaddr = local_handle.sectoraddr + local_handle.offset - dirent_size;
  WriteEEPromLong(directaddr, sectoraddr);
  FreeSectorChain(sectoraddr); // Free up all the blocks
  return 0;
}

// This routine creates a file
static long CreateFile(char *fname,  long modebits)
{
  long filesize;
  if (FindFile(fname, &local_handle, &filesize)) return 0;
  modebits = (0x10 | modebits) &attrib_mask;
  UpdateDirectoryEntry(&local_handle, 0, modebits, 0, fname);
  return 1;
}

static long OpenRead(char *fname, FileHandleT *handle)
{
  long directaddr, firstaddr, filesize;
  if (!(firstaddr = FindFile(fname, &local_handle, &filesize)))
    return -ENOENT;
  if (firstaddr & ATTR_DIRECTORY)
    return -EISDIR;
  directaddr = local_handle.sectoraddr + local_handle.offset - dirent_size;
  InitFileHandle(handle, directaddr, firstaddr, filesize, handle->buffer);
  return 0;
}

static long OpenAppend(char *fname, FileHandleT *handle)
{
  long directaddr, firstaddr, filesize;
  if ((firstaddr = FindFile(fname, &local_handle, &filesize)))
  {
    if (firstaddr & ATTR_DIRECTORY)
      return -EISDIR;
    directaddr = local_handle.sectoraddr + local_handle.offset - dirent_size;
    InitFileHandle(handle, directaddr, firstaddr, filesize, handle->buffer);
    eefs_hseek(handle, filesize);
  }
  else
  {
    UpdateDirectoryEntry(&local_handle, handle, 0x16, 0, fname);
  }
  return 0;
}

static long OpenWrite(char *fname, FileHandleT *handle)
{
  long sectoraddr, filesize;
  if ((sectoraddr = FindFile(fname, &local_handle, &filesize)))
  {
    if (sectoraddr & ATTR_DIRECTORY)
      return -EISDIR;
    FreeSectorChain(sectoraddr &sector_mask); // Free up all the blocks
    local_handle.position -= dirent_size;
    local_handle.offset -= dirent_size;
    UpdateDirectoryEntry(&local_handle, handle, 0x16, 0, 0);
  }
  else
  {
    UpdateDirectoryEntry(&local_handle, handle, 0x16, 0, fname);
  }
  return 0;
}

static long UpdateDirectoryEntry(FileHandleT *dirhand, FileHandleT *handle,  long firstaddr,  long filesize,  char *fname)
{
  long directaddr;
  eefs_hwrite(dirhand, (char *)&firstaddr, 4);
  eefs_hwrite(dirhand, (char *)&filesize, 4);
  directaddr = dirhand->sectoraddr + dirhand->offset - 8;
  if (fname)
  {
    eefs_hwrite(dirhand, fname, fname_size);
  }
  eefs_hclose(dirhand); // Force writes to happen
  if (handle)
  {
    InitFileHandle(handle, directaddr, firstaddr, filesize, handle->buffer);
  }
  return 0;
}

/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
